import sys
import functools

def normalize_url(url):
    """ ensure a url is "properly" constructed

    Remove trailing slashes
    Add http if http or https is not specified

    Parameters
    ----------
    url:  string
          A string to return as a url

    """
    if url.endswith('/'):
        url = url[:-1]
    if not url.startswith('http://') and not url.startswith('https://'):
        url = "http://{}".format(url)
    return url




# -*- coding: utf-8 -*-
import collections
import glob
import json
import os
import sys

# thirdparty libraies
import click
import requests

# This package
import ar_too as at
from ar_too.exceptions import UnknownArtifactoryRestError

# "CONSTANTS"
FETCH_WHAT_HELP_STR = """ What type of configs you want to fetch.  current options are repos (which takes optional --include_defaults and --include_filter arguements
"""
FETCH_REPO_TYPE_HELP_STR = """Artifactory repo type.  One of LOCAL, REMOTE, VIRTUAL.  If not given, all will be retreived."""

def _get_ldap_dict(ldap_json):
    """ return an OrderedDict for the given json file

    Parameters
    ----------
    ldap_json : string
        filepath to json file with config options to be loaded

    Returns
    -------
    ldap_dict : collections.OrderedDict
        ordered dictionary for use in configuring artifactory
    """
    try:
        with click.open_file(ldap_json) as f:
            json_dict = json.load(f, object_pairs_hook=collections.OrderedDict)
    except:
        click.echo("Can't open that LDAP json file")
        raise

    return json_dict

def _config_ldap(url, username, password, ldap_json):
    """ _config_ldap gets the current configuration and a json file, and
    update the config if necessary

    Parameters
    ----------
    url : string
        url for artifactory server
    username : string
        admin username on artifactory server
    password : string
        password for admin user
    ldap_json :
        filepath to json file the represents the ldap dictionary
    """
    auth = (username, password)
    current_conf = at.get_artifactory_config_from_url(url, auth)
    ldap_dict = _get_ldap_dict(ldap_json)

    new_conf, changed = at.update_ldapSettings_from_dict(current_conf, ldap_dict)
    if changed:
        click.echo("Modifying LDAP settings...")
        success = at.update_artifactory_config(url, auth, new_conf)
    else:
        click.echo("LDAP settings unchanged.")
        success = True

    if not success:
        click.echo("Something went wrong")
        sys.exit(1)

def _config_repos(url, username, password, repo_dir):
    """ for each file in the directory, create or update that repo

    Each file should be a json file of the format
    https://www.jfrog.com/confluence/display/RTF/Repository+Configuration+JSON

    Parameters
    ----------
    url : string
        url for artifactory server
    username : string
        admin username on artifactory server
    password : string
        password for admin user
    repo_dir : string
        path to a directory with repository config json files

    Notes
    -----
    This function will organize the repositories in two groups:
    first, local and remote repos
    second, virtual repos.

    This is because virtual repos aggregate local and remote repos and thus
    the locals and remotes must be present before we create the virtuals
    """

    repos_list_dict = _get_repos_from_directory(repo_dir)
    ses = requests.Session()
    ses.auth = (username, password)
    for rclass in ['local', 'remote', 'virtual']:
        for repo_dict in repos_list_dict[rclass]:
            success = at.cr_repository(url, repo_dict, session=ses)
            if success:
                click.echo("Successfully updated {}".format(repo_dict['key']))
            else:
                click.echo("Failed updating {}".format(repo_dict['key']))

def _get_repos_from_directory(repo_dir):
    """ return a dictionary of lists with 3 keys:
    local, remote, virtual.

    Each item of the list will be a dictionary representing one of these:
    https://www.jfrog.com/confluence/display/RTF/Repository+Configuration+JSON

    Parameters
    ----------
    repo_dir : string
        path to a directory with repository config json files.  see above

    Returns
    -------
    repos_list_dict : dictionary
        see description above

    Notes
    -----
    This will ONLY find .json files
    """

    if not os.path.isdir(repo_dir):
        click.echo("{} is not a directory.".format(repo_dir))
        sys.exit(1)

    repos_list_dict = {
            "local": [],
            "remote": [],
            "virtual": []
            }
    for jfile in glob.glob('{}/*.json'.format(repo_dir)):
        with click.open_file(jfile) as f:
            jdict = json.loads(f.read())

        try:
            rclass = jdict['rclass']
            repos_list_dict[rclass].append(jdict)
        except KeyError:
            click.echo("file {} as no rclass key.".format(jfile))
            click.echo("Skipping.")

    return repos_list_dict

def _config_admin_pass(host_url, password, target_password):
    """ set the admin password for Artifactory

    Parameters
    ----------
    host_url : string
        url for artifactory server, including the /artifactory context
    password : string
        password for admin user
    target_password : string
        desired password for admin user
    """
    try:
        changed = at.update_password(
                host_url,
                'admin',
                password,
                target_password
                )
    except UnknownArtifactoryRestError as ae:
        click.echo("Failed to update password.")
        click.echo(ae.msg)
        raise
    except:
        click.echo("Failed to update password for reasons unknown.")
        raise

    if changed:
        click.echo("Password successfully changed")
    else:
        click.echo("Password already at target")

def _fetch_repos(host_url, username, password, inc_defaults, inc_filter,
        output_dir, repo_type):
    """ download json configurations for repos, place them in output dir

    Parameters
    ----------
    host_url : string
        url for artifactory server, including the /artifactory context
    username : string
        admin username on artifactory server
    password : string
        password for admin user
    inc_defaults : boolean
        Whether we should include the repos artifactory ships with
    inc_filter : string
        Only include repos whose key includes this filter
    output_dir : string
        directory to write json files to
    repo_type : string
        One of the 3 artifactory repo types (LOCAL, REMOTE, VIRTUAL)
    """
    if repo_type is not None:
        repo_type = repo_type.upper()
        if repo_type not in ["LOCAL", "REMOTE", "VIRTUAL"]:
            click.echo("repo_type must be one of LOCAL, REMOTE, VIRTUAL")
            sys.exit(1)
    else:
        repo_type = "ALL"

    if not os.path.isdir(output_dir):
        click.echo("Can't find target directory. Exiting")
        sys.exit(1)

    repo_obj_list = at.get_repo_list(
            host_url,
            repo_type=repo_type,
            include_defaults=inc_defaults,
            include_filter=inc_filter
            )

    if len(repo_obj_list) == 0:
        click.echo("No repos found. Check your options")
        sys.exit(1)

    repo_list = [r['key'] for r in repo_obj_list]
    repo_config_list = at.get_repo_configs(
            host_url,
            repo_list,
            username=username,
            passwd=password
            )

    for repo in repo_config_list:
        repo_conf_file = os.path.join(output_dir, '{}.json'.format(repo['key']))
        with open(repo_conf_file, 'w') as f:
            f.write(json.dumps(repo, indent=4))

@click.group()
@click.option('--username', help="username with admin privileges")
@click.option('--password', help="password for user")
@click.option('--url', help="url and port for the artifactotry server")
@click.pass_context
def cli(ctx, **kwargs):
    """ Main entrypoint for ar_too cli """
    ctx.obj = kwargs

@cli.group()
@click.pass_context
def fetch(ctx, **kwargs):
    """ fetch command group """
    ctx.obj.update(kwargs)

@fetch.command()
@click.option('--include_defaults', is_flag=True, default=False,
        help="include artifactory default repos with the repos arg")
@click.option('--include_filter',
        help="only repos that include this in their key will be fetched")
@click.option('--output_dir', default=os.getcwd(),
        help="directory to place files")
@click.option('--repo_type', help=FETCH_REPO_TYPE_HELP_STR)
@click.pass_context
def repos(ctx, **kwargs):
    """ commands for retreiving configs from artifactory
    """
    ctx.obj.update(kwargs)

    _fetch_repos(
        ctx.obj['url'],
        ctx.obj['username'],
        ctx.obj['password'],
        ctx.obj['include_defaults'],
        ctx.obj['include_filter'],
        ctx.obj['output_dir'],
        ctx.obj['repo_type']
        )

@cli.command()
@click.option('--ldap_json', help="json file for ldap settings")
@click.option('--repos_dir', help="Dir with repository configuration files")
@click.option('--admin_pass', help="set new admin password to this")
@click.pass_context
def configure(ctx, **kwargs):
    """ command(s) for configuring artifactory
    """
    ctx.obj.update(kwargs)

    if ctx.obj['ldap_json'] is not None:
        _config_ldap(
            ctx.obj['url'],
            ctx.obj['username'],
            ctx.obj['password'],
            ctx.obj['ldap_json']
            )

    if ctx.obj['repos_dir'] is not None:
        _config_repos(
            ctx.obj['url'],
            ctx.obj['username'],
            ctx.obj['password'],
            ctx.obj['repos_dir']
            )

    if ctx.obj['admin_pass'] is not None:
        if ctx.obj['username'] != 'admin':
            click.echo("Must use the admin user to update the admin user")
            sys.exit(1)

        _config_admin_pass(
            ctx.obj['url'],
            ctx.obj['password'],
            ctx.obj['admin_pass']
            )



 # -*- coding: utf-8 -*-

class ConfigFetchError(Exception):
    """Raised when we were unable to fetch the config from the server
    """

    def __init__(self, msg, response):
        self.msg = msg
        self.response = response
        super(Exception, self)

class InvalidAPICallError(Exception):
    pass

class UnknownArtifactoryRestError(Exception):
    """ Raised if we don't know what happened with an artifactory rest call"""

    def __init__(self, msg, response):
        self.msg = msg
        self.response = response
        super(Exception, self)

class InvalidCredentialsError(Exception):
    """
    Raised if none of the provided crendentials provide the correct
    authorization
    """
    pass



# -*- coding: utf-8 -*-
import os
from copy import deepcopy

# thirdparty
import requests
import xmltodict

# this package
from utils import normalize_url
from exceptions import ConfigFetchError, InvalidAPICallError, InvalidCredentialsError, UnknownArtifactoryRestError

ART_REPO_TYPES = ["ALL", "LOCAL", "REMOTE", "VIRTUAL"]
ART_DEFAULT_REPOS = [
            'ext-release-local',
            'ext-snapshot-local',
            'libs-release-local',
            'libs-snapshot-local',
            'plugins-release-local',
            'plugins-snapshot-local',
            'jcenter-cache',
            'libs-release',
            'libs-snapshot',
            'plugins-release',
            'plugins-snapshot',
            'remote-repos',
            'jcenter'
        ]

def update_password(host_url, username, orig_pass, target_pass):
    """ set the password for the user to the target_pass

    Parameters
    ----------
    host_url : string
        A url of the form http(s)://domainname:port/context or
        http(s)://ip:port/context
    username : string
        username of the password to change
    orig_pass : string
        original password to use for the update
    target_pass : string
        the desired new password

    Returns
    -------
    changed : boolean
        True if changes were made

    Raises
    ------
    InvalidCredentialsError :
        If neither the original or target credentials work to update the password
    UnknownArtifactoryRestError :
        If we get a response we haven't encountered and don't know what to do with
    """

    orig_auth = (username, orig_pass)
    target_auth = (username, target_pass)

    get_pass_url = '{}/api/security/encryptedPassword'.format(
            normalize_url(host_url)
            )

    orig_resp = requests.get(get_pass_url, auth=orig_auth)
    if orig_resp.status_code == 401:
        resp = requests.get(get_pass_url, auth=target_auth)
        auth = target_auth
    elif orig_resp.status_code == 200:
        resp = orig_resp
        auth = orig_auth
    else:
        raise UnknownArtifactoryRestError(
                "Unexpected response when verifying credentials",
                orig_resp
                )

    if resp.status_code != 200:
        raise InvalidCrentialsError

    if auth == target_auth:
        return False

    user_json_url = '{}/api/security/users/{}'.format(
            normalize_url(host_url),
            username
            )

    headers = {'Content-type': 'application/json'}
    user_dict_resp = requests.get(user_json_url, auth=auth)
    if not user_dict_resp.ok:
        if user_dict_resp.status == 401:
            msg = "Received an unauthorized message after authorization "
            msg += "has been checked.  Wtf?"
            raise UnknownArtifactoryRestError(msg, user_dict_resp)
        else:
            raise UnknownArtifactoryRestError(
                    "Couldn't get user information",
                    user_dict_resp
                    )

    admin_dict = user_dict_resp.json()
    admin_dict.pop('lastLoggedIn')
    admin_dict.pop('realm')
    admin_dict['password'] = target_pass

    update_resp = requests.post(
            user_json_url,
            auth=auth,
            json=admin_dict,
            headers=headers
            )

    if not update_resp.ok:
        if update_resp.status == 401:
            msg = "Received an unauthorized message after authorization "
            msg += "has been checked.  Wtf?"
            raise UnknownArtifactoryRestError(msg, update_resp)
        else:
            raise UnknownArtifactoryRestError(
                    "Couldn't post user password update",
                    update_resp
                    )

    final_check_resp = requests.get(get_pass_url, auth=target_auth)
    if not final_check_resp.ok:
        raise UnknownArtifactoryRestError(
                "Final password check failed.  Could not use new credentials",
                final_check_resp
                )

    else:
        return True

def get_artifactory_config_from_url(host_url, auth):
    """retrieve the artifactory configuration xml doc

    Parameters
    ----------
    host_url:   string
        A url of the form http(s)://domainname:port/context or
        http(s)://ip:port/context
    auth:       tuple
                a tuple a la requests auth of the form (user, password)
    """
    headers = {'Accept': 'application/xml'}
    config_url = "{}/api/system/configuration".format(
            normalize_url(host_url)
            )

    r = requests.get(config_url, auth=auth, headers=headers)
    if r.ok:
        return(xmltodict.parse(r.text))
    else:
        raise ConfigFetchError("Something went wrong getting the config", r)

def update_ldapSettings_from_dict(config_dict, desired_dict):
    """match the ldap settings in the config_dict to the desired endstate

    Parameters
    ----------
    config_dict : dictionary
                 the source configuration dictionary
    desired_dict : dictionary
                  the ldap subdictionary that we want to use

    Returns
    -------
    return_dict : dictonary
        A copy of the original config dict, plus any modfications made
    changed : boolean
        Whether or not changes were made
    """

    return_dict = deepcopy(config_dict)
    orig_ldap_settings = return_dict['config']['security']['ldapSettings']
    if orig_ldap_settings == desired_dict:
        return return_dict, False

    # RED at the very least, this should validate the resulting xml
    # or, it should only update the changed keys so we know what they are
    # consider using easyXSD, but might want to avoid lxml
    else:
        return_dict['config']['security']['ldapSettings'] = desired_dict
        return return_dict, True

def update_artifactory_config(host_url, auth, config_dict):
    """ take a configuraiton dict and upload it to artifactory

    Parameters
    ----------
    host_url : string
        A url of the form http(s)://domainname:port[/context] or http(s)://ip:port[/context]
    auth : tuple
        A tuple of (user, password), as used by requests
    config_dict : OrderedDict
        a dict representation that will be returned to xml

    Returns:
    --------
    success : boolean
        true if we succeeded
    """
    headers = {'Content-type': 'application/xml'}
    config_url = "{}/api/system/configuration".format(
            normalize_url(host_url)
            )
    xml_config = xmltodict.unparse(config_dict)

    r = requests.post(config_url, auth=auth, headers=headers, data=xml_config)

    if r.ok:
        return True
    else:
        return False

def cr_repository(host_url, repo_dict, auth=None, session=None):
    """ take a configuration dict and post it host_url

    Should use
    https://www.jfrog.com/confluence/display/RTF/Repository+Configuration+JSON
    for the inputs.

    Does not error checking; will fail if the json is malformed.

    Parameters
    ----------
    host_url : string
        A url of the form
        http(s)://domainname:port[/context] or http(s)://ip:port[/context]
    repo_dict : OrderedDict
        a dictionary of the inputs required by artifactroy.  see above.
    auth : tuple, optional
        A tuple of (user, password), as used by requests
    session : requests Session object, optional
        A session object (that has any necessary cookies / headers defined)

    Either auth or session must be defined.  Session overrides auth.

    Returns
    -------
    success : boolean
        true if succeeded

    """
    ses = _get_artifactory_session(
            auth=auth,
            session=session
            )
    
    if 'key' not in repo_dict:
        raise InvalidAPICallError("The repo_dict must include a repo key (repo_dict['key'])")

    repo_url = '{}/api/repositories/{}'.format(
            normalize_url(host_url),
            repo_dict['key']
            )

    headers = {'Content-type': 'application/json'}
    exists_resp = ses.get(repo_url)
    if exists_resp.ok:
        resp = ses.post(repo_url, json=repo_dict, headers=headers)
    else:
        resp = ses.put(repo_url, json=repo_dict, headers=headers)

    # YELLOW need to add more logic to make this aware of if the configuration
    # is changing
    if resp.ok:
        return True
    else:
        return False

def _get_artifactory_session(username=None, passwd=None, auth=None,
        session=None):
    """ return a session with auth set.  prioritizes existing sessions,
        but validates that auth is set

    Parameters
    ----------
    username : string, optional
        username to create auth tuple from
    password : string, optional
        password for auth tuple
    auth : tuple, optional
        A tuple of (user, password), as used by requests
    session : requests.Session
        A requests.Session object, with auth

    Returns
    -------
    InvalidAPICallError
        When no combination of required inputs is given
    """
    if session is None and auth is None and username is None and passwd is None:
        raise InvalidAPICallError(
                "You must pass either username/password, auth, or session"
                )
    ses = None
    if session:
       if session.auth:
           ses = session
    if auth and not ses:
        ses = requests.Session()
        ses.auth = auth

    if (username and passwd) and not ses:
        auth = (username, passwd)
        ses = requests.Session()
        ses.auth = auth

    if not ses:
        raise InvalidAPICallError(
                "You must pass either username/password, auth, or session"
                )

    return ses

def get_repo_configs(host_url, repo_list, username=None, passwd=None,
        auth=None, session=None):
    """ return repository configuration dictionaries for specified set of repos

    Parameters
    ----------
    host_url : string
        An artifactory url of the form
        http(s)://domainname:port[/context] or http(s)://ip:port[/context]
    repo_list : list of strings
        A list of repo keys that you want to get configs for.  repo
        keys should match the url in the artifactory rest call
    username : string, optional
        username to create auth tuple from
    passwd : string, optional
        password for auth tuple
    auth : tuple, optional
        A tuple of (user, password), as used by requests
    session : requests.Session
        A requests.Session object, with auth

    Either session, auth, or user/pass must be defined.
    Session overrides auth overides username/password
    See _get_artifactory_session for details
    """
    ses = _get_artifactory_session(
            username=username,
            passwd=passwd,
            auth=auth,
            session=session
            )

    repo_configs_list = []
    for repo in repo_list:
        repo_url = '{}/api/repositories/{}'.format(
                normalize_url(host_url),
                repo
                )
        resp = ses.get(repo_url)
        if not resp.ok:
            msg = "Failed to fetch config for {}".format(repo)
            raise UnknownArtifactoryRestError(msg, resp)

        repo_dict = resp.json()
        repo_configs_list.append(repo_dict)

    return repo_configs_list


def get_repo_list(host_url, repo_type="ALL", include_defaults=False,
        include_filter=None):
    """ return repository configuration dictionaries for specified set of repos

    Parameters
    ----------
    host_url : string
        An artifactory url of the form
        http(s)://domainname:port[/context] or http(s)://ip:port[/context]
    repo_type : {'all', 'LOCAL', 'REMOTE', 'VIRTUAL'}
        What types of repo (as defined by artifactory) to fetch.
    include_defaults : boolean
        Whether to include repos that ship with artifactory
    include_filter : string
        String which is used to do a simple filter of repo names. (in)
        Using this + naming convention can filter by package type.

    Either session, auth, or user/pass must be defined.
    Session overrides auth overides username/password
    See _get_artifactory_session for details
    """
    repos_url = '{}/api/repositories'.format(host_url)
    resp = requests.get(repos_url)
    if not resp.ok:
        raise UnknownArtifactoryRestError("Error fetching repos", resp)

    final_repo_list = resp.json()

    if repo_type.upper() != "ALL":
        if repo_type.upper() not in ART_REPO_TYPES:
            raise InvalidAPICallError("repo_type must be one of {}".format(
                ART_REPO_TYPES
                )
            )
        final_repo_list = [r for r in final_repo_list
            if r['type'] == repo_type.upper()]

    if include_filter:
        final_repo_list = [r for r in final_repo_list
            if include_filter in r['url'].split('/')[-1]]

    if not include_defaults:
        final_repo_list = [r for r in final_repo_list
            if r['key'] not in ART_DEFAULT_REPOS]

    return final_repo_list


if __name__ == '__main__':
    print("This file is not the entrypoint for ar_too")
    sys.exit(1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup

with open('README.rst') as readme_file:
    readme = readme_file.read()

requirements = [
    "click",
    "requests",
    "xmltodict"
]

test_requirements = [
    "click",
    "mock",
    "pytest",
    "xmltodict"
]

setup_requirements = []

setup(
    name='ar_too',
    version="0.5",
    use_scm_version=True,
    setup_requires=setup_requirements,
    description="Artifactory configuration tool",
    long_description=readme,
    author="MCP CI",
    author_email='mcp-ci@mirantis.com',
    url='https://github.com/brain461/ar_too.git',
    download_url = 'https://github.com/brain461/ar_too/tarball/0.4',
    packages=['ar_too'],
    package_dir={'ar_too':
                 'ar_too'},
    include_package_data=True,
    install_requires=requirements,
    license="Apache",
    zip_safe=False,
    keywords='ar_too',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    entry_points={
        'console_scripts': ['ar_too=ar_too.cli:cli']
    }
)
